
:- ensure_loaded( library(lists) ). % member/2, memberchk/2.
:- ensure_loaded( '../../src/init_lib' ).

:- requires( opts_select_infiles/3 ).

report_terms_with_matching_values( Opts ) :-
	opts_select_infiles( Opts, Fs, NonFOpts ),
	findall( Val, member(val(Val),NonFOpts), Vals ),
	( Vals == [] -> 
		throw( at_least_one_option(val/1)
		        -required_for(report_terms_with_matching_values/1) )
		;
		true
	),
	( memberchk(pos(Pos),NonFOpts) -> 
		true
		;
		throw( option_required( pos/1 )
		        -required_for(report_terms_with_matching_values/1) )
	),
	( memberchk(out(Out),NonFOpts) -> true ; Out = line ),
	( memberchk(op(InOp),NonFOpts) -> 
		once( operator_to_prolog( InOp, Op ) )
		;
		Op = (==)
	),
	report_terms_matching( Fs, Vals, Pos, Op, Out ).

report_terms_matching( [], _Vals, _Pos, _Op, _Out ).
report_terms_matching( [H|T], Vals, Pos, Op, Out ) :-
	open( H, read, In ),
	read( In, Term ),
	write( file-H ), nl,
	write( vals-Vals ), nl, 
	report_terms_matching_on_stream( Term, 1, In, Vals, Pos, Op, Out ),
	close( In ),
	report_terms_matching( T, Vals, Pos, Op, Out ).

report_terms_matching_on_stream( end_of_file, _Ln, _In, _Vals, _Pos, _Op, _Out ) :-
	!.
report_terms_matching_on_stream( InTerm, InLn, In, Vals, Pos, Op, Out ) :-
	arg( Pos, InTerm, ArgAtPos ),
	( arg_matches_some_val(Vals,Op,ArgAtPos) -> 
		report_matched_term( Out, InTerm, ArgAtPos, InLn ), nl
		;
		% write( nm(ArgAtPos) ), nl,
		true
	),
	NxLn is InLn + 1,
	read( In, NxTerm ),
	report_terms_matching_on_stream( NxTerm, NxLn, In, Vals, Pos, Op, Out ).

report_matched_term( [H|T], InTerm, ArgAtPos, InLn ) :-
	!,
	report_matched_term( H, InTerm, ArgAtPos, InLn ),
	write( ' | ' ), nl,
	report_matched_term( T, InTerm, ArgAtPos, InLn ).

report_matched_term( line, _InTerm, _ArgAtPos, InLn ) :-
	!,
	write( InLn ).
report_matched_term( term, InTerm, _ArgAtPos, _InLn ) :-
	!,
	write( InTerm ).
report_matched_term( pos, _InTerm, ArgAtPos, _InLn ) :-
	!,
	write( ArgAtPos ).
report_matched_term( all, InTerm, ArgAtPos, InLn ) :-
	write( InLn ), write( ' | ' ),
	write( InTerm ), write( ' | ' ),
	write( ArgAtPos ).

arg_matches_some_val( [H|_T], Op, Arg ) :-
	functor( Call, Op, 2 ),
	arg( 1, Call, H ),
	arg( 2, Call, Arg ),
	call( Call ),
	!.
arg_matches_some_val( [_H|T], Op, Arg ) :-
	arg_matches_some_val( T, Op, Arg ).

operator_to_prolog( leq, (@>=) ).
operator_to_prolog( le, (@>) ).
operator_to_prolog( geq, (@=<) ).
operator_to_prolog( gr, (@<) ).
operator_to_prolog( eq, (==) ).
operator_to_prolog( ar_leq, (>=) ).
operator_to_prolog( ar_le, (>) ).
operator_to_prolog( ar_geq, (=<) ).
operator_to_prolog( ar_gr, (<) ).
operator_to_prolog( ar_eq, (=:=) ).
