
:- ensure_loaded( '../src/init_lib' ).				% library alias.
:- ensure_loaded( '../src/all_dynamic'  ).			% /1
:- ensure_loaded( '../src/ad_to_slp'  ).			% /1.

:- ensure_loaded( library(lists) ).				% memberchk/2.

:- requires( opts_have_help/1 ).
:- requires( slp_in_options/2 ).
:- requires( pl_to_slp_call/0 ).	% _args/3.

enum_models( Opts ) :-
	opts_have_help( Opts ),
	!, nl,
	write( 'enum_models       enumerate and record all possible models generated' ), nl,
	write( '                  by an SLP. ' ), nl, nl,
	write( ' Options:' ), nl,
	write( '       slp/1      the SLP file' ), nl,
	write( '       pl/1       alternative to SLP, use Prolog file' ), nl,
	write( '       call/2     call and observable' ), nl,
	write( '       post/1     name of predicate with arity two. Operation to perform' ), nl,
	write( '                  on each Observable. Def is =' ), nl,
	write( '       rec/1      name of predicate with arity/2. If absent m(Obs,[Post])' ), nl,
	write( '                  will be recorded' ), nl,
	write( '       out/1      output file.' ), nl, nl.

enum_models( Opts ) :-
	memberchk( call(PlCall,Obs), Opts ),
	( memberchk(slp(_),Opts) -> 
		slp_in_options( Opts, Slp ),
		all_dynamic( Slp ),
		ad_to_slp( [tmp(tmp_file),rm(false)] ),
		Mod = slp,
		pl_to_slp_call_args( PlCall, TheCall, _CallArgs )
		;
		( memberchk(pl(Pl),Opts) -> 
			ensure_loaded(Pl),
			Mod = user,
			TheCall = PlCall
			;
			write( user_error, 'At least one of slp/1, pl/1 is needed.' ),
			nl( user_error ),
			abort
		)
	),
	( memberchk( post(Post),Opts) -> true ; Post = (=) ),
	( memberchk( rec(Rec),Opts) -> true ; 
		( Post = (=) -> Rec = write_m_obs_only ; Rec = write_m ) ),
	memberchk( out(Out), Opts ),
	open( Out, write, OutStream ),
	current_output( Curr ),
	set_output( OutStream ),
	bb_put( streams, s(Curr,OutStream) ),
	call(Mod:TheCall),
	PostCall =.. [Post,Obs,Sec],
	call(PostCall),
	RecCall =.. [Rec,Obs,Sec],
	call(RecCall),
	fail.
enum_models( _ ) :-
	bb_get( streams, s(Std,File) ),
	set_output( Std ),
	close( File ).

write_m( Obs, Post ) :- write( m(Obs,Post) ), write( '.' ), nl.
write_m_obs_only( Obs, _ ) :- write( m(Obs) ), write( '.' ), nl.
