:- ensure_loaded( '../src/init_lib' ).		%.
:- ensure_loaded( library(lists) ).	% member/2, sum_list/2,
								% append/3, max_list/2.
:- ensure_loaded( library(system) ).	% mktemp/2, system/1.
:- ensure_loaded( library(charsio) ).	% write_to_chars/2

% :- multifile( library_directory/1 ).
% :- dynamic( library_directory/1 ).
% 
% library_directory( lib ).
% library_directory( '../../src/lib' ).

:- ensure_loaded( library(check_list) ).
:- ensure_loaded( library(gnucomfile) ).		%.
:- ensure_loaded( library(flatoms) ).		%.
:- ensure_loaded( library(delete_file_if) ).		%.
:- ensure_loaded( library(kv_decompose) ).		%.
% :- ensure_loaded( library(copy_stream_on_stream) ).	% /2.
:- ensure_loaded( library(file_to_list_of_lines) ).		%.
:- ensure_loaded( library(shorten_long_lines) ).		% /5.
:- ensure_loaded( library(list_of_lines_to_file) ).	% /2.
:- ensure_loaded( library(replace_all_sublists ) ).	% /4.
% :- pl( yap(_), ensure_loaded( library(max_list_yap) ) ).

% Stats = X-Y pairs, Stem to be used for generated files
% From, Col, Program, Transform.
%
eps_n_latex_graph_freqs( Stats, Stem, From, Col, Program, Transform ) :-
	enl_gf_defaults( Defs ),
	eps_n_latex_graph_freqs( Stats, Stem, From, Col, Program, Transform, Defs ).
eps_n_latex_graph_freqs( Stats, Stem, From, Col, Program, Transform, Opts ) :-
	enl_gf_defaults( Defs ),
	append( Opts, Defs, All ),
	eps_graph_freqs( Stats, Stem, Eps, Opts ),
	latex_eps_report_on_freqs( Stats, Eps, Stem, From, Col, _Latex, Program, Transform, All ),
	atom_concat( Stem, '.ps', Ps ),
	( memberchk(gv(true),All) -> 
		atom_concat( Ps, ' & ', PsApp ),
		atom_concat( 'gv ', PsApp, Gv ),
		write( Gv ), 
		system( Gv ),
		nl, flush_output
		;
		write( see_files(Eps,Ps)), nl
	).

enl_gf_defaults( [tabular(true),package(graphicx)] ).

% Stats is a list of frequency pairs Obs-Times.
% Times, is the frequency of Obs in Stats.
% This predicate (a) generates and gv an eps graph of Times vs Obserbations
% and (b) produces a latex report that includes the graph and a table of
% the frequencies. PrvEps is the location of the eps file. It can be a variable
% in which case a uniquie file of the form 'eps_*" will be used. A '.eps' will
% be added if the provided file doesnt have this extention. The name of the
% file used is in Eps.
% 
eps_graph_freqs( Stats, PrvEps, Eps, LtxOpts ) :-
	mktemp( freq_histo_XXXXXX, DataFile ),
	open( DataFile, write, Stream ),
	write( Stream, 0 ), nl( Stream ),
	Goal = (Pair=_K-V,write(Stream,V),nl(Stream)),
	check_list( Stats, Goal, Pair ),
	close( Stream ),
	kv_decompose( Stats, _Obs, Vs ),
	max_list( Vs, ThatMany ),
	% sum_list( Vs, ThatMany ),
	give_space( ThatMany, ThatHigh ),
	length( Stats, Length ),
	give_space( Length, ThatWide ),
	mktemp( gnuplot_coms_XXXXXX, GnuFile ),
	( var(PrvEps) -> mktemp( eps_XXXXXX, PrvEps ) ; true ),
	( atom_concat( _Stem, '.eps', PrvEps ) -> 
		Eps = PrvEps 
		;
		atom_concat( PrvEps, '.eps', Eps )
	),
	% flatoms( ['output "',Eps,'"'], EpsOpt ),
	( ThatWide > 20 ->  Xmin is -0.5 ; Xmin is 0.5 ),
	( memberchk(with(With),LtxOpts) -> true ; With = boxes ),
	Opts = [ % set('boxwidth 1'),
			term('postscript eps'),
			output(Eps),with(With),x(Xmin,ThatWide),y(0,ThatHigh) ],
	gnucomfile( GnuFile, Opts, DataFile ),
	flatoms( ['gnuplot \"',GnuFile,'\"'], Gplot ),
	system( Gplot ),
	sleep( 2 ),
	delete_file_if( GnuFile ),
	delete_file_if( DataFile ),
	delete_file_if( PrvEps ).

give_space( ThatMany, ThatHigh ) :-
	PrvThatHigh is ThatMany + ThatMany // 10,
	( PrvThatHigh =:= ThatMany ->
		( PrvThatHigh > 4 -> 
			ThatHigh is ThatMany + 1
			;
			ThatHigh is ThatMany + (ThatMany / 3.1415)
		)
		;
		ThatHigh is PrvThatHigh
	).
latex_eps_report_on_freqs( Stats, Eps, PrvTex, From, Col, Tex, Prg, Tra, Opts ) :-
	( var(PrvTex) -> 
		mktemp( report_XXXXXX, PrvTex )
		; 
		true
	),
	( atom_concat( _StemTex, '.tex', PrvTex ) -> 
		Tex = PrvTex 
		;
		atom_concat( PrvTex, '.tex', Tex )
	),	
	open( Tex, write, Stream ),
	latex_out( Stream, Stats, Eps, From, Col, Prg, Tra, Opts ),
	close( Stream ),
	% flatoms( ['latex ',Tex], Latex ),
	% flatoms( ['latex ',Tex], Latex ),
	flatoms( ['latex ',Tex,' >& /dev/null '], Latex ),
	system( Latex ),
	system( Latex ),
	atom_concat( PrvTex, '.dvi', Dvi ),
	atom_concat( PrvTex, '.ps', Ps ),
	flatoms( ['dvips -o ',Ps,' ',Dvi,' >& /dev/null'], DviPs ), 
	system( DviPs ),
	delete_file_if( PrvTex ).

latex_out( S, Stats, Eps, From, Col, Prg, Tra, Opts ) :-
	memberchk( tabular(Tblr), Opts ),
	FigRef = thefiglabel,
	% TblRef = thetbllabel,
	write( S, '\\documentclass{article}' ), nl( S ), nl( S ),
	latex_use_packages( Opts, S ),
	write( S, '\\title{Predicate frequencies}' ), nl( S ), nl( S ),
	write( S, '\\begin{document}' ), nl( S ), nl( S ),
	write( S, '\\maketitle' ), nl( S ), nl( S ),
	write( S, '\\section{Some statistics}' ), nl( S ),
	write( S, '   \\begin{figure}[htbp]' ), nl( S ),
	write( S, '     \\includegraphics*{' ), write( S, Eps ),
	write( S, '}' ), nl( S ),
	write( S, '\\label{' ), write( S, FigRef ),
	write( S, '}' ), nl( S ),
	write( S, 'Fig.~\\ref{' ), write( S, FigRef ), write( S, '}.' ),
	write( S, 'Automatically generated statistics for ' ),
	write( S, From ), nl( S ),
	write( S, '   \\end{figure}' ), nl( S ),
	nl( S ), nl( S ),
	% write( S, '   \\begin{table}[htbp]' ), nl( S ),
	write( S, '   \\section{The Full Frequencies}' ), nl( S ),
	columns_to_latex_lls( Col, [], LLs ),
	( Tblr == false -> 
		true 
		;
		write( S, '     \\begin{tabular}{' ), write( S, LLs ),
		 write( S, '}' ), nl( S )
	),
	write_list_per_n_elements( Stats, Tblr, 1, Col, S ),
	% write_list_per_two_elements( Stats, S ),
	( Tblr == false -> 
		true
		; write( S, '     \\end{tabular}' ), nl( S ) ),
	% write( S, '\\label{' ), write( S, TblRef ), write( S, '}' ), nl( S ),
	% write( S, '\\caption{' ),
	% write( S, 'Table~\\ref{' ), write( S, FigRef ), write( S, '}' ),
	% write( S, 'The full frequencies.' ),
	% write( S, '}' ), nl( S ),
	% write( S, '   \\end{table}' ), nl( S ),
	nl( S ),
	% write( S, 'See Fig.~\\ref{' ),
	% write( S, FigRef ), write( S, '} and Table~\\ref{' ),
	% write( S, TblRef ), write( S, '}' ),
	% nl( S ),
	( var(Prg) ->
		true
		; 
		write( S, '   \\section{SLP Program}' ), nl( S ),
		% write( S, '   \\begin{table}[htbp]' ), nl( S ),
		write( S, '     \\begin{verbatim}' ), nl( S ),
		% open( Prg, read, In ),
		file_to_list_of_lines( Prg, InPrgLns ),
		replace_all_sublists_nested( InPrgLns, "\t","        ", PrgLns ),
		shorten_long_lines( PrgLns, 80, "", 10, "        ", OkLns ),
		list_of_lines_to_stream( OkLns, S ),
		% copy_stream_on_stream( In, S ),
		% close( In ),
		write( S, '     \\end{verbatim}' ), nl( S ),
		% write( S, '\\label{' ), write( S, TblRef ), write( S, '}' ), nl( S ),
		% write( S, '\\caption{' ),
		% write( S, 'The program.' ), write( S, '}' ), nl( S ),
		% write( S, '   \\end{table}' ), nl( S )
		nl( S )
	),
	( var(Tra) -> true 
		;
		write( S, '   \\section{Transformed Program}' ), nl( S ),
		% write( S, '   \\begin{table}[htbp]' ), nl( S ),
		write( S, '     \\begin{verbatim}' ), nl( S ),
		file_to_list_of_lines( Tra, TraLns ),
		shorten_long_lines( TraLns, 80, "", 10, "        ", TraOkLns ),
		list_of_lines_to_stream( TraOkLns, S ),
		% open( Tra, read, InOth ),
		% copy_stream_on_stream( InOth, S ),
		% close( InOth ),
		write( S, '     \\end{verbatim}' ), nl( S ),
		% write( S, '\\label{' ), write( S, TblRef ), write( S, '}' ), nl( S ),
		% write( S, '\\caption{' ),
		% write( S, 'Transformed program.' ), write( S, '}' ), nl( S ),
		% write( S, '   \\end{table}' ), nl( S )
		nl( S )
	),
	% write( S, ), nl( S ),
	% write( S, ), nl( S ),
	% write( S, ), nl( S ),
	% write( S, ), nl( S ),
	% write( S, ), nl( S ),
	% write( S, ), nl( S ),
	write( S, '\\end{document} ' ).

latex_use_packages( Opts, S ) :-
	write( S, '\\usepackage{' ),
	findall( P, member(package(P),Opts), Ps ),
	latex_write_use_packages( Ps, S ).

latex_write_use_packages( [], S ) :- write( S, '}' ), nl( S ).
latex_write_use_packages( [H|T], S ) :-
	write( S, H ),
	( T = [_|_] -> write( S, ',' ) ; true ),
	latex_write_use_packages( T, S ).
	% write( S, '\\usepackage{graphicx,float} ' ), nl( S ), nl( S ),
	

write_src_files( [], S ) :- 
	write( S, 'no file found.' ).
write_src_files( [H], S ) :-
	!,
	write( S, 'file ' ),
	write_protected_underscore( H, S ).
write_src_files( [H|T], S ) :-
	write( S, 'files ' ),
	write_protected_underscore( H, S ),
	write_tail_separated( T, S, ', ', ' and ', '.' ).

write_tail_separated( [], _S, _, _, _ ).
write_tail_separated( [H|T], S, Btw, LstB, Fin ) :-
	( T = [] -> 
		write( S, LstB ),
		write_protected_underscore( H, S ),
		write( S, Fin )
		;
		write( S, Btw ),
		write_protected_underscore( H, S ),
		write_tail_separated( T, S, Btw, LstB, Fin )
	).

write_list_per_n_elements( [], _Tblr, _K, _N, S ) :- nl( S ).
write_list_per_n_elements( [H], _Tblr, _K, _N, S ) :- 
	!,
	write_stat_elem_to( H, S ), nl( S ).
write_list_per_n_elements( [H|T], Tblr, K, N, S ) :- 
	write_stat_elem_to( H, S ),
	( Tblr == false ->
		true
		;
		( 0 =:= K mod N  -> 
			write( S, ' \\\\ ' ), nl( S )
			; 
			write( S, ' & ' )
		)
	),
	NxK is K + 1,
	write_list_per_n_elements( T, Tblr, NxK, N, S ).

write_stat_elem_to( Name/Arity-Times, S ) :-
	!,
	write_protected_underscore( Name, S ), 
	write( S, '/' ), write( S, Arity ),
	write( S, ' - ' ), write( S, Times ).
write_stat_elem_to( Item-Times, S ) :-
	( bb_get(tex_model,(PName,Inp)) -> 
		Call =.. [PName,Item,S,Inp,CleanUp],
		call( Call )
		;
		CleanUp = true
	),
	write_to_chars( Item-Times, Chars ),
	shorten_long_lines( [Chars], 80, "'", 10, "        ", OkChars ),
	add_tex_nl_in_between_cs_lines( OkChars, TexLines ),
	write( S, '\'' ),
	this_list_of_lines_to_stream( TexLines, S ),
	% write( S, Item-Times ),
	write( S, ' \'' ), 
	call( CleanUp ).

protect_underscore( [], [] ).
protect_underscore( [0'_|T], [0'\\,0'_|R] ) :-
	!,
	protect_underscore( T, R ).
protect_underscore( [H|T], [H|R] ) :-
	protect_underscore( T, R ).

add_tex_nl_in_between_cs_lines( [F,S|T], [FNL|TNL] ) :-
	!,
	append( F, " \\\\ ", FNL ),
	add_tex_nl_in_between_cs_lines( [S|T], TNL ).
add_tex_nl_in_between_cs_lines( SorE, SorE ).

write_protected_underscore( Name, S ) :-
	atom_codes( Name, NmCs ),
	protect_underscore( NmCs, PrctCs ),
	atom_codes( Prct, PrctCs ),
	write( S, Prct ).

% predi( [src('predi.pl')] ).
% cn '/homes/nicos/pl/programs/predi'.
% compile( '/homes/nicos/pl/programs/predi' ).
% predi( [src('bg.pl'),src('unrelated_egs.slp')] ).
%

columns_to_latex_lls( 1, Acc, LLs ) :-
	!,
	atom_codes( LLs, [0'l|Acc] ).
columns_to_latex_lls( N, Acc, LLs ) :-
	NxN is N - 1,
	columns_to_latex_lls( NxN, [0'l|Acc], LLs ).

replace_all_sublists_nested( [], _This, _With, [] ).
replace_all_sublists_nested( [L|Ls], This, With, [F|Rs] ) :-
	replace_all_sublists( L, This, With, F ),
	replace_all_sublists_nested( Ls, This, With, Rs ).

this_list_of_lines_to_stream( [], _S ).
this_list_of_lines_to_stream( [H|T], S ) :-
	this_fput_line( H, S ),
	( T == [] -> true ; nl( S ) ),
	this_list_of_lines_to_stream( T, S ).

this_fput_line( [], _S  ).
this_fput_line( [C|Cs], S ) :-
        put( S, C ),
        this_fput_line( Cs, S ).
