:- ensure_loaded( '../src/init_lib' ).		%.
:- ensure_loaded( library(directory_contents) ).		% /4.
:- ensure_loaded( library(file_to_list_of_lines) ).	% /2.
:- ensure_loaded( library(kv_transpose) ).			% /2.

:- ensure_loaded( library(lists) ).				% append/3, select/3.
:- ensure_loaded( library(system) ).				% working_directory/2.

% In = [disp(Disp),sort(Sort)|Dirs]
% order is not important but maximum one for disp/1 and sort/1 is allowed.
% Disp in {times,*both*}, Sort in {*none*,lex}.
%
timings( In ) :-
	( select(sort(Sort),In,NxIn) ->
		true
		;
		Sort = none,
		NxIn = In
	),
	( select(disp(Disp),NxIn,Files) ->
		true
		;
		Disp = both,
		Files = NxIn
	),
	Opts = [sort(Sort),disp(Disp)],
	timings_1( Files, Opts, [] ).

timings_1( [], [sort(Sort),disp(Disp)], All ) :-
	once( sort_timings(Sort,All,Sorted) ), 
	record_timings( Sorted, Disp ).

timings_1( [H|T], Opts, Acc ) :-
	dir_timings( H, These ),
	append( Acc, These, NxAcc ),
	timings_1( T, Opts, NxAcc ).

dir_timings( H, HTimings ) :-
	working_directory( Old, H ),
	directory_contents( '.', Files, _Dirs, _Others ),
	select_stats_files( Files, StatsFiles ),
	extract_timings_from_stats_files( StatsFiles, HTimings ),
	working_directory( _Hag, Old ).

extract_timings_from_stats_files( [], [] ).
extract_timings_from_stats_files( [H|T], [Htim-H|Ttims] ):-
	file_to_list_of_lines( H, Lines ),
	last( Lines, Last ),
	number_codes( Secs, Last ),
	Htim is Secs / 3600,
	extract_timings_from_stats_files( T, Ttims ).

select_stats_files( [], [] ).
select_stats_files( [H|T], Stats ) :-
	( atom_concat(_,'.stats',H) ->
		Stats = [H|TStats]
		;
		Stats = TStats
	),
	select_stats_files( T, TStats ).

record_timings( [], _ ).
record_timings( [Tim-File|T], Disp ) :-
	format( "~t~16g~8|", [Tim] ),
	( Disp == both ->
		write( ' - ' ), write( File ), nl
		;
		nl
	),
	record_timings( T, Disp ).

sort_timings( none, Sorted, Sorted ).
sort_timings( lex, All, Sorted ) :-
	kv_transpose( All, TransAll ), 
	keysort( TransAll, TransSorted ),
	kv_transpose( TransSorted, Sorted ).
