:- ensure_loaded( rpdag ). 				% /2.
:- ensure_loaded( library(lists ) ).		% append/3, memberchk/2.
:- ensure_loaded( library(ugraphs) ).		% edges/2 add_edges/3, transpose/2,
									% add_vertices/3.
:- ensure_loaded( '../auxil/bn_to_eps' ).


disp_rpdags( Opts ) :-
	% Nodes  = [1,2,3],
	( memberchk(nodes(NodesTok),Opts) ->
		nodes_token( NodesTok, Nodes )
		; Nodes  = [1,2,3,4] ),
	findall( RpDag, (rpdag(Nodes,RpDag),nl), RpDags ),
	% check_dirs( RpDags, 1 ), halt,
	disp_rpdag_defs( Nodes, Defs ),
	append( Opts, Defs, CompletedOpts ),
	memberchk( disp(Disp), CompletedOpts ),
	memberchk( repo(Repo), CompletedOpts ),
	once( disp_rpdags_1( Disp, Repo, Nodes, RpDags, _ ) ).

check_dirs( [], _I ).
check_dirs( [rpdag(Pairs,Dir,Und)|T], I ) :-
	( top_sort(Dir,_Sort) -> write( ok(I) )
		; nl, write( cyclic(rpdag(Pairs,Dir,Und)) ), nl ),
	NxI is I + 1,
	check_dirs( T, NxI ).

nodes_token( three, [1,2,3] ).
nodes_token( four, [1,2,3,4] ).
nodes_token( five, [1,2,3,4,5] ).

disp_rpdag_defs( Nodes, [disp(tex),repo(Stem)] ) :-
	nodes_to_stem( Nodes, Stem ).

nodes_to_stem( [_,_,_], three ).
nodes_to_stem( [_,_,_,_], four ).

% disp_rpdags_1( eps, Nodes, RpDags ).
% disp_rpdags_1( tex, Nodes, RpDags ).
% disp_rpdags_1( dot, Nodes, RpDags ).
% disp_rpdags_1( viz, Nodes, RpDags ).

disp_rpdags_1( eps, Repo, Nodes, RpDags, Files ) :-
	disp_rpdags_1_eps( RpDags, 1, Repo, Nodes, Files ).
disp_rpdags_1( tex, Repo, Nodes, RpDags, _ ) :-
	fname_extension( Repo, ".tex", Stem, ExtRepo, _ ),
	disp_rpdags_1( eps, Stem, Nodes, RpDags, Files ),
	tell( ExtRepo ),
	texcln( 'documentclass{article}' ), nl,
	texcln( 'usepackage{graphicx,amsmath,float}' ), nl,
	length( Nodes, Length ), % number_codes( 3, LCodes ),
	% atom_codes( LAtom, LCodes ),
	write( '\\title{All RPDAGs for ' ), write( Length ),
	write( ' Nodes.}' ), nl, nl,
	texcln( 'begin{document}' ),
	texcln( maketitle ), nl,
	texcln( 'begin{figure}[H]' ),
	texcln( 'begin{center}' ),
	figures_for_epses( Files, 1, 4 ),
	texcln( 'end{document}' ),
	told.

disp_rpdags_1_eps( [], _, _, _, [] ).
disp_rpdags_1_eps( [rpdag(_,Dir,Und)|T], I, Stem, Nodes, [HF|TFs] ) :-
	edges( Dir, DirEdges ),
	add_edges( Und, DirEdges, Full ),
	number_codes( I, ICs ), atom_codes( IAtm, ICs ),
	atom_concat( '_', IAtm, UscAtm ), atom_concat( Stem, UscAtm, HF ),
	transpose( Full, BnLike ),
	add_vertices( BnLike, Nodes, BnFull ),
	bn_to_eps( BnFull, HF, [delete(false),cline('-Gnodesep=0.1 -Nfontsize=8 -Nwidth=0.3 -Nheight=0.15')] ),
	NxI is I + 1,
	disp_rpdags_1_eps( T, NxI, Stem, Nodes, TFs ).

figures_for_epses( [], _I, _N ) :-
	nl, texcln( 'end{center}' ), texcln( 'end{figure}' ).
figures_for_epses( [H|T], I, N ) :-
	( I > N -> 
		nl, texcln( 'end{center}' ),
		texcln( 'end{figure}' ), nl, NxI is 2,
		texcln( 'begin{figure}[H]' ),
		texcln( 'begin{center}' )
		;
		NxI is I + 1
	),
	write( ' \\includegraphics*{' ), 
	write( H ),
	( I =:= N -> write( '}'); write( '}\\hspace{0.3cm}' ) ),
	figures_for_epses( T, NxI, N ).

texcln( Atom ) :-
	write( '\\' ), write( Atom ), nl.
