:- ensure_loaded( library(lists) ).		% select/3, append/3.
:- ensure_loaded( library(ugraphs) ).		% del_edges/3, add_edges/3.
:- ensure_loaded( library(kv_decompose) ).	% /3.

rpdag_to_bn( rpdag(Pairs,Dir,Und), Bn ) :-
	rpdag_links_to_directed( Und, Dir, ConnGr ),
	kv_decompose( Pairs, Nodes, _ ),
	add_vertices( ConnGr, Nodes, Gr ),
	% RPDAG components are in library format, V-Children
	% BNs   are in V-Parents format. Thus:
	transpose( Gr, Bn ).

rpdag_links_to_directed( [], Dir, Dir ).
rpdag_links_to_directed( [V-Chln|T], Acc, Dir ) :-
	direct_children_links( Chln, V, T, Acc, [], MdUnd, MdAcc, Away ),
	direct_away_from_nodes( Away, MdUnd, MdAcc, RmUnd, NxAcc ),
	rpdag_links_to_directed( RmUnd, NxAcc, Dir ).

direct_children_links( [], _V, Und, Dir, Away, Und, Dir, Away ).
direct_children_links( [H|T], V, AccUnd, AccDir, AccAway, Und, Dir, Away ) :-
	del_edges( AccUnd, [H-V], MdUnd ),
	add_edges( AccDir, [V-H], MdDir ),
	( select(H-HChln,MdUnd,NoHUnd) ->
		direct_away_from( HChln, H, NoHUnd, MdDir, NxUnd, NxDir )
		; 
		write( user_error, something_funny(H,MdUnd) ), nl( user_error ),
		abort
	),
	append( AccAway, HChln, NxAway ),
	direct_children_links( T, V, NxUnd, NxDir, NxAway, Und, Dir, Away ).

direct_away_from_nodes( [], Und, Dir, Und, Dir ).
direct_away_from_nodes( [H|T], InUnd, InDir, Und, Dir ) :-
	( select(H-HChln,InUnd,NoHUnd) ->
		direct_away_from( HChln, H, NoHUnd, InDir, NxUnd, NxDir )
		; 
		write( user_error, something_strange(H,InUnd) ), nl( user_error ),
		abort
	),
	direct_away_from_nodes( T, NxUnd, NxDir, Und, Dir ).

direct_away_from( [], _V, Und, Dir, Und, Dir ).
direct_away_from( [H|T], V, InUnd, InDir, Und, Dir ) :-
	del_edges( InUnd, [H-V], NxUnd ),
	add_edges( InDir, [V-H], NxDir ),
	direct_away_from( T, V, NxUnd, NxDir, Und, Dir ).

% add_disconnected_nodes_to_graph( [], ConnBn, Gr ).
% add_disconnected_nodes_to_graph( [H-_|T], ConnBn, Gr ) :-
	% add_disconnected_nodes_to_graph( T, ConnBn, Gr ).
