
/* 2007, November. see classify_unknown( help ). */

:- ensure_loaded( library(system) ). % /1, environ/2.

:- ensure_loaded( '../../src/init_lib' ).
:- ensure_loaded( 'classify_unknown_w_chain_in_memory' ). % /1.

:- requires( opts_have_help/2 ).
:- requires( write_help/2 ).
:- requires( dir_will_exist/1 ).
:- requires( directory_contents/4 ).
:- requires( file_stems_of_files_with_this_ext/3 ).
:- requires( slashify/2 ).
:- requires( flatoms/2 ).
:- requires( basename/2 ).
:- requires( to_list/1 ).

classify_unknown( Opts ) :-
     opts_have_help( Opts ),
     !,
     classify_unknown_help( Opts ).

classify_unknown( PrvOpts ) :-
     to_list( PrvOpts, Opts ),
     classify_unknown_defaults( Defs ),
     append( Opts, Defs, All ),
     memberchk( unknowns_dir(Unk), All ),
     memberchk( descriptors(Dsc), All ),
     ( memberchk(null_value(_),Opts) ->
          NullOpts = Opts; NullOpts = Defs ),
     findall( Null, (member(null_value(Null),NullOpts),
                 assert(null_value(Null)) ), _Nulls ),
     directory_contents( Unk, Files, _SubDirs1, _Other1 ),
     file_stems_of_files_with_this_ext( Files, '.txt', Stems ),
     slashify( Unk, UnkSl ),
     construct_unknown_file_locations( Stems, UnkSl, Unknowns ),
     memberchk( chains_dir(Chn), All ),
     directory_contents( Chn, LocalFs, _SubDirs2, _Other2 ),
     dir_will_exist( unk_clsf ),
     file_stems_of_files_with_this_ext( LocalFs, '.fterms', FtermStems ),
     slashify( Chn, ChnSl ),
     classify_unknown_1( FtermStems, ChnSl, Unknowns, Dsc ).

classify_unknown_1( [], _Dir, _Unknowns, _Dsc ).
classify_unknown_1( [H|T], Dir,  Unknowns, Dsc ) :-
     write( doing_fterm_stem(H) ), nl,
     atom_concat( Dir, H, DirH ),
     atom_concat( DirH, '.fterms', Hfterms ),
     classify_unknown_w_chain_in_memory( [Hfterms,Dsc|Unknowns] ),
     % atom_concat( 'unk_clsf/', H, LocalStem ),
     % move_unknown_classified_files_to_stemmed( Unknowns, LocalStem ),
     classify_unknown_1( T, Dir, Unknowns, Dsc ).

construct_unknown_file_locations( [], _UnkSl, [] ).
% construct_unknown_file_locations( [], _UnkSl, _Unknowns ).
construct_unknown_file_locations( [H|T], Dir, [F|M] ) :-
     atom_concat( H, '.txt', Hpl ),
     atom_concat( Dir, Hpl, F ),
     construct_unknown_file_locations( T, Dir, M ).

move_unknown_classified_files_to_stemmed( [], _LocalStem ).
move_unknown_classified_files_to_stemmed( [H|T], LocalStem ) :-
     atom_concat( Hstem, '.txt', H ),
     basename( Hstem, Hbase ),
     flatoms( ['mv ', Hstem, '_cls.pl ', LocalStem, '_', Hbase], Mv ),
     shell( Mv ),
     move_unknown_classified_files_to_stemmed( T, LocalStem ).

classify_unknown_defaults( Defs ) :-
     environ( 'HOST', Host ),
     atom_concat( '$HOME/res/ed2_descs/ky1571_100k/text_do_', Host, Dir ),
     Defs = [
          % unknowns_dir('.'),
          unknowns_dir(Dir),
          chains_dir('.'),
          descriptors(1571),
          null_value(-999)
               ].

classify_unknown_help( _HelpOpts ) :-
     classify_unknown_defaults( Defs ),
     % ( memberchk(unknowns_dir(Lim),Defs)-> true; Lim = 1000),
     Descr  = 'classify a number of feature files according to each fterm file in a directory',
     Opts = [
               opt(unknowns_dir,dir,'all .txt files in this dir are considered feature values to be classified, '),
               opt(chains_dir,dir,'all .fterms files in this dir are considered to be from mcmc chains, '),
               opt(descriptors,dsc,'number of descriptor variables, '),
               opt(null_value,null,'null value, ')
     ],
     Exs = 'lupsh classify_unknown % from fterms dir',
     write_help( help(f(classify_unknown),Descr,Opts,Exs), Defs ).

