
/*
     Try to load the classifiers for all models of a single run into memory
     and then use these to classify a number of examples in input files.

     */

:- ensure_loaded( library(lists) ).          % memberchk/2.
:- ensure_loaded( '../../src/init_lib' ). 
:- elm( 'carts/src/cart_classifies' ).       % cart_to_classifier/4.
:- elm( 'carts/src/cart_classifies_unknown' ).
                   % cart_to_classifier/3, cart_classifier_add_dominant_class/2.
:- elm( 'src/model_type_calls' ).            % /1.

:- ensure_loaded( library(matrix) ).         % matrix_new/3.
% :- ensure_loaded( '$HOME/pl/lib/term/delim_line_to_tree' ).

:- requires( spc_delim_nums_line_to_matrix/4 ).
:- requires( read_terms/2 ).
:- requires( dirpath_will_exist/1 ).
:- requires( flatoms/2 ).

classify_unknown_w_chain_in_memory( [Chain,Dsc|Files] ) :-
     open( Chain, read, In ),
     read( In, Term ),
     atom_concat( Stem, '.fterms', Chain ),
     atom_concat( Stem, '.stats', Stats ),
     basename( Stem, BaseStem ),
     read_terms( Stats, FstTerms ),
     memberchk( data_file(DFile), FstTerms ),
     memberchk( data_call(DCall), FstTerms ),
     memberchk( model_type(MdlType), FstTerms ),
     % setrand( rand(1,1,1) ),
     model_type_calls( MdlType ),
     cart_data_call_and_file( DCall, DFile, Ids ),
     chain_classifiers_to_memory( Term, Ids, In, 1, Classifiers ),
     close( In ),
     write( read_classifiers_from(Chain) ), nl,
     shell( date ),
     !,
     classify_unknown_w_chain_in_memory_1( Files, BaseStem, Classifiers, Dsc ),
     shell( date ).

chain_classifiers_to_memory( end_of_file, _Ids, _In, _I, [] ).
chain_classifiers_to_memory( st(Tms,_Lhd,_Min,Cart), Ids, In, I, [Tms-Csf|TCsfs] ) :-
     ( (I mod 10000) =:= 0 -> write( i(I) ), nl; true ),
     cart_to_classifier( Cart, Ids, bayes, CartCsf ),
     cart_classifier_add_dominant_class( CartCsf, Csf ),
     read( In, Nxt ),
     NxtI is I + 1,
     chain_classifiers_to_memory( Nxt, Ids, In, NxtI, TCsfs ).
     
cart_data_call_and_file( idsd_fprs(RelTrain,Ids,_), DFile, Ids ) :-
     write( user_error, ensure_loaded( DFile ) ), nl( user_error ),
     ensure_loaded( DFile ),
     call( idsd_fprs(RelTrain,ids(Ids,_),_) ).

classify_unknown_w_chain_in_memory_1( [], _Fstem,  _Classifiers, _Dsc ).
classify_unknown_w_chain_in_memory_1( [H|T], Fstem, Classifiers, Dsc ) :-
     write( data_h(H) ), nl,
     open( H, read, In ),
     atom_concat( Stem, '.txt', H ),
     % atom_concat( Stem, '_cls.pl', Hcls ),
     basename( Stem, Base ),
     dirpath_will_exist( unk_clsf ),
     flatoms( ['unk_clsf/',Fstem,'_',Base,'.pl'], Hcls ),
     open( Hcls, write, Out ),
     % Classifiers = [_Htms-Hclsf|_],
     % cart_classifies_unknown( Hclsf, Fst, _Class-Distro ),
     % distro_to_null_acc( Distro, Null ),
     Null = [0,0], % Here, need to fix this for more than 2 classes....
     Lim is Dsc + 1,
     matrix_new( floats, [1,Lim], Mtx ),
     classify_unknowns_on_stream_w_chain_in_memory( Mtx, Dsc, Classifiers, Null, In, Out ),
     close( In ),
     close( Out ),
     !,
     classify_unknown_w_chain_in_memory_1( T, Fstem, Classifiers, Dsc ).

classify_unknowns_on_stream_w_chain_in_memory( _Mtx, _Dsc, _Classifiers, _Null, In, _Out ) :-
     % use end_of_file, when reading lists with values in. 
     % this version uses a single row matrix which is read on the fly
     at_end_of_stream( In ),
     !.
classify_unknowns_on_stream_w_chain_in_memory( Mtx, Dsc, Clsfs, Null, In, Out ) :-
     %%  delim_line_to_tree( In, Tree ), %% this is faster than reading-in
     %%  the data list, but slower than a single matrix. 
     %%  i havent tried to use a single tree.
     spc_delim_nums_line_to_matrix( Dsc, 1, In, Mtx ),
     classifiers_on_unknown( Clsfs, 0, Mtx, Null, Null, Classes, Res ),
     portray_clause( Out, cas(Classes,Res) ),
     % read( In, Nxt ),
     classify_unknowns_on_stream_w_chain_in_memory( Mtx, Dsc, Clsfs, Null, In, Out ).

classifiers_on_unknown( [], _Seen, _Vals, Classes, Prbs, Classes, Prbs ).
classifiers_on_unknown( [Tms-H|T], Seen, Vals, AccC, AccP, Classes, Prbs ) :-
     cart_classifies_unknown( H, Vals, NClass-Distro ),
     nth_addition( NClass, AccC, Tms, NxtC ),
     add_new_probs( AccP, Distro, Seen, Tms, NxtP ),
     NxSeen is Tms + Seen,
     classifiers_on_unknown( T, NxSeen, Vals, NxtC, NxtP, Classes, Prbs ).

add_new_probs( [], [], _Seen, _Tms, [] ).
add_new_probs( [Av|Avs], [H|T], Seen, Tms, [F|M] ) :-
     F is Av + ( (H-Av) * Tms / (Seen+Tms) ),
     add_new_probs( Avs, T, Seen, Tms, M ).

distro_to_null_acc( [], [] ).
distro_to_null_acc( [_|T], [0|TNull] ) :-
     distro_to_null_acc( T, TNull ).

nth_addition( 1, [Old|Tail], Add, [New|Tail] ) :- !,
     New is Old + Add.
nth_addition( I, [H|Tail], Add, [H|More] ) :-
     NxI is I - 1,
     nth_addition( NxI, Tail, Add, More ).
