:- ensure_loaded( library(charsio) ).		% write_to_chars/2.
:- ensure_loaded( library(system) ).		% file_exists/1, system/1, sleep/1,
									% working_directory/2.
:- ensure_loaded( library(lists) ).		% memberchk/2....
:- ensure_loaded( '../../src/init_lib' ).			% library().

:- ensure_loaded( lld_stats_for_stem ).				% lld_stats_for_stem_1/1.

:- requires( opts_select_infiles/3 ).

:- ensure_loaded( library(directory_contents) ).		%.
:- ensure_loaded( library(directory_files_nodots) ).	% /2.
:- ensure_loaded( library(file_stems_of_files_with_this_or_no_ext) ).
:- ensure_loaded( library(mktemp_in_dir) ).             % /3.
:- ensure_loaded( library(gnucomfile) ).		% 3.
:- ensure_loaded( library(sys_info) ).		% /1
:- ensure_loaded( library(contains) ).		% /2.
:- ensure_loaded( library(file_to_list_of_lines) ).	% /2.
:- ensure_loaded( library(mcmcms_top_dir) ).	% /2.
:- ensure_loaded( library(elm) ).	% /1.
:- ensure_loaded( bests_in_fterms_file ).	% /3.
% :- ensure_loaded( library(parse_number) ).	% /3.
:- elm( 'carts/src/display_model' ).		% /1.
:- elm( 'auxil/dataset_elm_location' ).		% /2.
:- elm( 'carts/auxil/red_cart_to_full' ).	% /4.

:- prolog_load_context(directory,Dir), atom_concat( Dir, '/', DirSl ),
   bb_put( report_src_directory, DirSl ).

eps_outs( Epses ) :-
	Epses = [ stays, visits, depths, leaves, llhoods,
	          mst_visited, bst_llhood ].

% Option bnw(true) to produce displayed models in b&w too.
% true is default.
report( Opts ) :-
	write( user_error, single_data_set_assumed ),
	nl( user_error ),
	opts_select_infiles( Opts, PrvDirs, RmOpts ),
	( memberchk(set(Set),RmOpts) ->
		ensure_loaded_data_set( Set, NFts )
		;
		true
	),
	% file_dir_1_or_atom( Opts, PrvDirs, RmOpts ),
	( PrvDirs == [] -> Dirs = ['.'] ; Dirs = PrvDirs ),
	report_dirs( Dirs, RmOpts, Set, NFts ).

report_dirs( [], _RmOpts, _Set, _NFts ).
report_dirs( [Dir|TDirs], RmOpts, Set, NFts ) :-
	% add following to options
	( bb_get(refresh,_Bool) -> true ; bb_put(refresh,false) ),
	( DirOr = [Dir] -> true ; (DirOr = [] -> Dir = '.' ; Dir = DirOr )),
	working_directory( Old, Dir ),
	bb_put( reporting_in_dir, Old ),
	directory_contents( '.', _Files, AllSubDirs, _Oth ),
	select_rrun_subdirs( AllSubDirs, USubDirs ),
	sort( USubDirs, SubDirs ),
	atom_concat( Dir, '/', DirSl ),
	report_exp_dirs( SubDirs, DirSl, RmOpts, Set, NFts, Index ),
	working_directory( _DirA, Old ),
	( file_exists('preIndex.tex') ->
		true
		;
		sys_info( 'cp $HOME/islp/mcmcms/carts/auxil/preIndex.tex .' )
	),
	sys_info( 'cp preIndex.tex Index.ltx' ), 
	open( 'Index.ltx', append, Out ),
	% write( Out, '\\documentclass{book}' ), nl( Out ),
	% write( Out, '\\usepackage{graphicx}' ), nl( Out ), nl( Out ),
	% write( Out, '\\begin{document}' ), nl( Out ), nl( Out ),
	eps_outs( Epses ),
	report_tex( Index, Epses, Out ),
	nl( Out ), write( Out, '\\end{document}' ), nl( Out ),
	close( Out ),
	sleep( 2 ),
	( (system( 'texps Index > /dev/null' ),system( 'texps Index > /dev/null' )) -> true ; true ),
	system( 'mkvis * >& /dev/null' ),
	system( 'gv Index.ps &' ),
	report_dirs( TDirs, RmOpts, Set, NFts ).

report_tex( [], _Eps, _Out ).
report_tex( [H|T], Eps, Out ) :-
	H = (HDir,HepsF),
	select_all_eps_in_dir( T, HDir, Matches, Rm ),
	atom_codes( HDir, HDirCs ),
	protect_underscore( HDirCs, PctHDirCs ),
	atom_codes( PctHDir, PctHDirCs ),
	write( Out, '\\chapter{' ), write( Out, PctHDir ), 
	write( Out, '}' ), nl( Out ),
	report_stems( [HepsF|Matches], HDir, Eps, Out ),
	% report_stem( Eps, Stem, Out ),
	% report_similarity( HDir, Out ),
	report_tex( Rm, Eps, Out ).

report_similarity( Dir, Out ) :-
	directory_files_nodots( Dir, Files ),
	atom_concat( Dir, '/', DirSl ),
	choose_sim_files_and_concat_to_dir( Files, DirSl, Epses ),
	write( Out, '\\section{Seeds comparisons}' ), nl( Out ),
	report_similarity_epses( Epses, Out ).

choose_sim_files_and_concat_to_dir( [], _Dir, [] ).
choose_sim_files_and_concat_to_dir( [H|T], Dir, Epses ) :-
	( atom_concat(_,'_sim.eps',H) -> 
		atom_concat( Dir, H, DirH ),
		Epses = [DirH|TEpses]
		;
		TEpses = Epses
	),
	choose_sim_files_and_concat_to_dir( T, Dir, TEpses ).

report_similarity_epses( [], _Out ).
report_similarity_epses( [EpsF|T], Out ) :-
	atom_codes( EpsF, EpsFCs ),
	protect_underscore( EpsFCs, PctEpsFCs ),
	atom_codes( PctEpsF, PctEpsFCs ),
	include_figure( EpsF, seed_compare, PctEpsF, Out ),
	report_similarity_epses( T, Out ).

report_stems( [], _Dir, _Eps, _Out ).
report_stems( [H|T], Dir, Eps, Out )  :-
	( atom_concat(Dir,Base,H) -> true 
		; Base = H ), % this should never be the case
	report_stem( Eps, H, Base, Out ),
	report_accuracy_in_tex( H, Base, Out ),
	report_stems( T, Dir, Eps, Out ).

report_stem( [], _Stem, _Base, _Out ).
report_stem( [H|T], Stem, Base, Out ) :-
	atom_concat( H, '.eps', Heps ),
	atom_concat( '_', Heps, UscHeps ),
	atom_concat( Stem, UscHeps, FullEps ),
	atom_concat( Base, UscHeps, FullBas ),
	report_eps_in_tex( FullEps, FullBas, H, Out ),
	% write( Out, '\\clearpage' ), nl( Out ),
	report_stem( T, Stem, Base, Out ).

report_eps_in_tex( EpsF, Base, Type, Out ) :-
	write_section( Base, Out ),
	atom_codes( EpsF, EpsFCs ),
	protect_underscore( EpsFCs, PctEpsFCs ),
	atom_codes( PctEpsF, PctEpsFCs ),
	include_figure( EpsF, Type, PctEpsF, Out ).

include_figure( EpsF, Type, Caption, Out ) :-
	write( Out, '\\begin{figure}[H]' ), nl( Out ),
	write( Out, '\\begin{center}' ), nl( Out ),
	write( Out, '    \\includegraphics*[width=1\\textwidth]{' ),
	write( Out, EpsF ), write( Out, '}' ), nl( Out ),
	write( Out, '\\end{center}' ), nl( Out ),
	write( Out, '\\caption{' ),
	once( caption_intro(Type,Out) ), write(Out,Caption),
	write( Out, '}' ), nl( Out ),
	write( Out, '\\end{figure}' ),
	nl( Out ).

report_accuracy_in_tex( H, Base, Out ) :-
	atom_concat( 'Accuracy: ', Base, Section ),
	% write_section( 'Accuracy', Out ),
	write_section( Section, Out ),
	report_in_tex_first_term_in_ext_as( H, '.accuracy_bin', 'Accuracy', Out ),
	nl( Out ), write( Out, '\\noindent ' ),
	report_in_tex_first_term_in_ext_as( H, '.accuracy_exp', 'Expected accuracy', Out ).

write_section( Name, Out ) :-
	write( Out, '\\section{' ),
	atom_codes( Name, NameCs ), 
	protect_underscore( NameCs, PctNameCs ),
	atom_codes( PctName, PctNameCs ),
	write( Out, PctName ), write( Out, '}' ), nl( Out).

report_in_tex_first_term_in_ext_as( Stem, Ext, As, Out ) :-
	atom_concat( Stem, Ext, File ),
	( file_exists(File) -> 
		open( File, read, In ), 
		read( In, BinTerm ),
		write( Out, As ), write( Out, ': ' ), 
		write( Out, BinTerm  ), write( Out, '.' ), nl( Out ),
		close( In )
		;
		write( user_error, acc_file_not(File) ), nl( user_error )
	).

select_all_eps_in_dir( [], _Dir, [], [] ).
select_all_eps_in_dir( [(Dir,File)|T], MatchDir, Matched, Rm ) :-
	( Dir == MatchDir ->
		Matched = [File|Tmatched],
		Rm = TRm
		;
		Matched = Tmatched,
		Rm = [(Dir,File)|TRm]
	),
	select_all_eps_in_dir( T, MatchDir, Tmatched, TRm ).

caption_intro( stays, Out ) :-
	write( Out, 'Stays on a model before jumping versus models, ' ).
caption_intro( visits, Out ) :-
	write( Out, 'Total number of visiting a distinct model, including returns, versus models, ' ).
caption_intro( depths, Out ) :-
	write( Out, 'Max depth versus transitions, ' ).
caption_intro( leaves, Out ) :-
	write( Out, 'Number of leaves versus transitions, ' ).
caption_intro( llhoods, Out ) :-
	write( Out, 'Log-likelihood versus transitions, ' ).
caption_intro( bst_llhood, Out ) :-
	write( Out, 'Model with best log-likelihood, ' ).
caption_intro( mst_visited, Out ) :-
	write( Out, 'Model visited the most, ' ).
caption_intro( seed_compare, Out ) :-
	write( Out, 'Seed comparison, ' ).

report_exp_dirs( [], _Dir, _Opts, _Set, _NFts, [] ).
report_exp_dirs( [H|T], Dir, Opts, Set, NFts, Index ) :-
	working_directory( Old, H ),
	write( entering(H) ), nl,
	directory_contents( '.', Files, _SubSubDirs, _Oth ),
	file_stems_of_files_with_this_or_no_ext( Files, '.gz', Stems ),
	atom_concat( H, '/', HSl ),
	atom_concat( Dir, HSl, FullDir ),
	report_experiments( Stems, FullDir, Set, NFts, Opts, Index/Tindex ),
	working_directory( _HA, Old ),
	write( back_to_top_directory ), nl,
	report_exp_dirs( T, Dir, Opts, Set, NFts, Tindex ).

report_experiments( [], _Dir, _Set, _NFts,  _Opts, Idx/Idx ).
report_experiments( [Hstem|Tstems], Dir, Set, NFts, Opts, Idx/Tin ) :-
	all_stats_for_exp_present( Hstem, Set, NFts ),
	atom_concat( Dir, Hstem, HIdx ),
	Idx = [(Dir,HIdx)|Ndx],
	mktemp_in_dir( gnuplotcoms_XXXXXX, '.', TmpComs ),
	report_eps( [visits,stays,llhoods,leaves,depths], Hstem, TmpComs ),
	report_best( Hstem, Dir, Opts, Set, NFts ),
	delete_file( TmpComs ),
	report_experiments( Tstems, Dir, Set, NFts, Opts, Ndx/Tin ).

report_eps( [], _Stem, _Tmp ).
report_eps( [Hext|Texts], Stem, Tmp ) :-
	atom_concat( Hext, '.eps', HextEps ),
	atom_concat( '_', HextEps, UscExt ),
	atom_concat( Stem, UscExt, HEps ),
	( (file_exists(HEps),bb_get(refresh,false)) -> 
		% change to bb_get(refresh,extension)
		true
		;
		atom_concat( '.', Hext, DotExt ),
		atom_concat( Stem, DotExt, Source ),
		atom_concat( Tmp, '\"', TmpDQ ),
		% atom_concat( Source, '\"', SourceSl ),
		atom_concat( 'gnuplot \"', TmpDQ, Gplot1 ),
		Disp = term('postscript eps'), 
		File = output(HEps),
		With = with(lines),
		gnucomfile( Tmp, [Disp,File,With], Source ),
		% exec( Gplot1, [null,null,null], _Pid2 )
		% sys_info( Gplot1 )
		( system(Gplot1) -> 
			write( ok(gnuplotting,Source) ), nl
			;
			write( failed(gnuplotting,Source) ), nl
		)
	),
	report_eps( Texts, Stem, Tmp ).
	
report_best( Stem, _Dir, _Opts, _DSet, _NFts ) :-
	atom_concat( Stem, '_mst_visited.eps', MstVis ),
	atom_concat( Stem, '_bst_llhood.eps', BstLhd ),
	file_exists( MstVis ),
	file_exists( BstLhd ),
	!.

report_best( Stem, Dir, Opts, DSet, NFts ) :-
	atom_concat( Stem, '.fterms', Fterms ),
	bests_in_fterms_file( Fterms, BTmsST, BLLST ),
	% atom_concat( ' -o ', Stem, OtMinStem ),
	atom_concat( Stem, '_mst_visited.eps', TmsOut ),
	atom_concat( Stem, '_bst_llhood.eps', LLOut ),
	BTmsST = st(BTms,BTLL,BTMis,PrvBTM),
	ensure_full_model( PrvBTM, BTM ),
	write_to_chars( most_vst(BTms):msclf(BTMis), BTCodes ),
	atom_codes( BTAtm, BTCodes ),
	atom_concat( Stem, '.classes', ClFile ),
	( file_exists(ClFile) -> bb_put( classes_file, ClFile ) ; true ),
	NVars is NFts + 1,
	( (current_predicate(data_hdr/NVars),
	   functor(Dhcall,data_hdr,NVars),
	   call(Dhcall)                    ) -> 
	   	Dhcall =.. [data_hdr|Hdr],
		hdr_add_ids( Hdr, 1, IdsHdr ),
		Add = [node_label(IdsHdr)]
		;
		Add = []
	),
	write( add(Add) ),
	display_model( BTM, [title(BTAtm),lhood(BTLL),exec(eps),out(TmsOut)|Add] ),
	BLLST = st(BLlTms,BLl,BLlMis,PrvBLlM),
	ensure_full_model( PrvBLlM, BLlM ),
	write_to_chars( best_llhood:vst(BLlTms):msclf(BLlMis), LlCodes ),
	atom_codes( LlAtm, LlCodes ),
	display_model( BLlM, [title(LlAtm),lhood(BLl),exec(eps),out(LLOut)|Add] ),
	( memberchk(bnw(false),Opts) -> 
		true
		;
		atom_concat( Stem, '_mst_visited_bnw.eps', BnWTmsOut ),
		atom_concat( Stem, '_bst_llhood_bnw.eps', BnWLLOut ),
		display_model( BTM, [title(BTAtm),lhood(BTLL),exec(eps),out(BnWTmsOut),colour(bnw),node_style(solid)|Add] ),
		% display_model( BTM, [title(BTAtm),lhood(BTLL),exec(eps),out(BnWTmsOut),colour(bnw),node_style(solid),node_label(excl_id)] ),
		% display_model( BLlM, [title(LlAtm),lhood(BLl),exec(eps),out(BnWLLOut),colour(bnw),node_style(solid),node_label(excl_id)] )
		display_model( BLlM, [title(LlAtm),lhood(BLl),exec(eps),out(BnWLLOut),colour(bnw),node_style(solid)|Add] )
	),
	( bb_delete( classes_file, _ ) -> true ; true ),
	atom_concat( Stem, '.stats', StatsFile ),
	atom_concat( 'cat ', StatsFile, Cat ),
	atom_concat( Cat, ' | grep datime > tmp_dtimes.pl', Grep ),
	system( Grep ),
	file_to_list_of_lines( StatsFile, StatsLines ),
	penultimate( StatsLines, PenultimateLn ),
	% parse_number( PenultimateLn, SecsCs, _ ),
	number_codes( Secs, PenultimateLn ),
	open( 'tmp_dtimes.pl', read, DTTmp ),
	read(  DTTmp, DTFrom ),
	read(  DTTmp, DTUntil ),
	close( DTTmp ),
	delete_file( 'tmp_dtimes.pl' ),
	atom_concat( Dir, Stem, DirStem ),
	open( '../best.pl', append, BestRecOut ),
	writeq( BestRecOut, best_of(DirStem,BLl,Secs,DTFrom-DTUntil,DSet) ),
	write( BestRecOut, '.' ), nl( BestRecOut ),
	close( BestRecOut ).

all_stats_for_exp_present( Stem, Set, NFts ) :-	
	( var(Set) ->
		stem_to_data_set( Stem, Set ),
		write( using_dset(Set) ), nl,
		ensure_loaded_data_set( Set, NFts )
		; true ),
	atom_concat( Stem, '.fterms', Fterms ),
	( file_exists(Fterms) ->
		true
		;
		atom_concat( Set, ' ', SetSpc ),
		atom_concat( 'upsh p fterms ', SetSpc, UpshPfx ),
		atom_concat( UpshPfx, Stem, Upsh ),
		sys_info( Upsh )
	),
	lld_stats_for_stem_1( Stem ),
	gzip_if_stem( Stem ).

stem_to_data_set( _InStem, FullSet ) :-
	bb_get( reporting_in_dir, There ),
	atom_codes( There, ThereCs ),
	write( there(There) ), nl,
	contains( ThereCs, "/ijcai05" ),
	FullSet = pima_synth_500,
	!.
stem_to_data_set( InStem, FullSet ) :-
	bb_get( reporting_in_dir, There ),
	atom_codes( There, ThereCs ),
	( contains( ThereCs, "/icml05/" ) ->
		PfxSet = 'icml_',
		working_directory( Stem, Stem )
		;
		PfxSet = '',
		Stem = InStem
	),
	atom_codes( Stem, StemCs ),
	( contains( StemCs, "kyph" ) ->
		Set = kyph
		; 
		( contains( StemCs, "bcw" ) -> 
			Set = bcw
			; 
			( contains( StemCs, "pima" ) -> 
				Set = pima
				;
				( contains(StemCs,"_sn") ->
					( append(_,[0's,0'n|RstStemCs],StemCs) ->
						( RstStemCs=[NumCd,0't|_] ->
							NumCs = [NumCd]
							;
							RstStemCs = [N1Cd,N2Cd|_],
							NumCs = [N1Cd,N2Cd]
						),
						append( "sn_clean", NumCs, SnNumCs ),
						( contains(StemCs,"sn12")->
							PsfxCs = "_trn_100"
							;
							( contains(StemCs,"sn13")->
								PsfxCs = "_trn_100"
								;
								PsfxCs = "_trn_080"
							)
						),
						% ( append( _, "tt", StemCs ) ->
							% PsfxCs = "trn_080"
							% ;
							% PsfxCs = ""
						% ),
						append( SnNumCs, PsfxCs, SnCleanCs ),
						atom_codes( Set, SnCleanCs )
						;
						write( user_error, not_a_known_stem(Stem) ),
						nl( user_error ),
						abort
					)
				)
			)
		)
	),
	atom_concat( PfxSet, Set, FullSet ),
	!.

protect_underscore( [], [] ).
protect_underscore( [0'_|T], [0'\\,0'_|R] ) :-
	!,
	protect_underscore( T, R ).
protect_underscore( [H|T], [H|R] ) :-
	protect_underscore( T, R ).

select_rrun_subdirs( [], [] ).
select_rrun_subdirs( [H|T], RRunDirs ) :-
	( atom_concat(rr,_,H) ->
		RRunDirs = [H|TRRunDirs]
		;
		( atom_concat(b,_,H) ->
			RRunDirs = [H|TRRunDirs]
			;
			RRunDirs = TRRunDirs
		)
	),
	select_rrun_subdirs( T, TRRunDirs ).

gzip_if_stem( Stem ) :-
	( file_exists(Stem) -> 
		% atom_concat( Stem, '.gz', GzFile ),
		atom_concat( 'gzip ', Stem, Gzip ),
		sys_info( Gzip )
		;
		true
	).

ensure_loaded_data_set( DSet, NFts ) :-
	( load_data_set(DSet,NFts)  ->
		true
		;
		throw( unknown_data_set(DSet) )
	).

load_data_set( Dset, NofFeats ) :-
	atom_concat( 'icml_', Which, Dset ), 
	!,
	bb_get( report_src_directory, Dir ),
	atom_concat( Which, '_train_080', Train ),
	atom_concat( Dir, '../icml05/data/', DataDir ),
	atom_concat( DataDir, Train, DFile ),
	atom_concat( Dir, '../icml05/data/idsd', IdsdF ),
	write( user_error, loading(DFile) ), nl( user_error ),
	ensure_loaded( DFile ),
	ensure_loaded( IdsdF ),
	memberchk( Which-NofFeats, [bcw-9,kyph-3,pima-8] ),
	idsd( NofFeats, Train, _Splits, _Ids ).
	% pima_data( _, _ ).
load_data_set( pima, 0 ) :-
	elm( 'carts/pima/data/pima_data' ),
	pima_data( _, _ ).
load_data_set( pima_synth_500, 8 ) :-
	elm( 'pbl/ijcai05/data/idsd' ),
	idsd( 8, sg500, _, _ ).
load_data_set( Sn, SnFts ) :-
	atom_concat( 'sn', _, Sn ),
	!,
	dataset_elm_location( sn_clean1, _Type, SnClean1 ),
	atom_concat( Base, 'sn_clean1.pl', SnClean1 ),
	atom_concat( Base, 'idsd.pl', Idsd ),
	elm( Idsd ),
	write( elm( Idsd ) ), nl,
	elm( 'auxil/dset_has_nofeatures' ),
	atom_concat( Base, Sn, BaseSn ),
	bb_get( mcmcms_top_dir, MTD ),
	atom_concat( MTD, BaseSn, FullSn ),
	dset_has_nofeatures( FullSn, SnFts ),
	/*
	Pairs = [sn_clean2_trn_080-19,sn_clean3_trn_080-33,
	         sn_clean4_trn_080-8, sn_clean5_trn_080-15],
	memberchk( Sn-SnFts, Pairs ),
	*/
	idsd( SnFts, Sn, _, _ ).

/*
load_data_set( sn_clean1_trn ) :-
	dataset_elm_location( sn_clean1_trn, _Type, SnClean1 ),
	atom_concat( Base, 'sn_clean1_trn_080.pl', SnClean1 ),
	atom_concat( Base, 'idsd.pl', Idsd ),
	elm( Idsd ),
	idsd( 18, sn_clean1_trn_080, _, _ ).
load_data_set( sn_clean1 ) :-
	dataset_elm_location( sn_clean1, _Type, SnClean1 ),
	atom_concat( Base, 'sn_clean1.pl', SnClean1 ),
	atom_concat( Base, 'idsd.pl', Idsd ),
	elm( Idsd ),
	idsd( 18, sn_clean1, _, _ ).
*/

penultimate( [F,S|T], Penultimate ) :-
	penultimate_1( T, F, S, Penultimate ).

penultimate_1( [], Penultimate, _Last, Penultimate ).
penultimate_1( [H|T], _, Lst, Penultimate ) :-
	penultimate_1( T, Lst, H, Penultimate ).

ensure_full_model( PrvMod, Mod ) :-
	( reduced_cart_model(PrvMod) ->
		bb_get( idsd_put, t(_NofF,_DataSet,_Splits,Ids) ),
		red_cart_to_full( PrvMod, 0, Ids, Mod )
		;
		Mod = PrvMod
	).

hdr_add_ids( [], _N, [] ).
hdr_add_ids( [H|T], N, [N-H|R] ) :-
	NxN is N + 1,
	% atom_codes( H, HCs ),
	% hdr_to_lower( HCs, HlowCs ),
	% atom_codes( Hlow , HlowCs ),
	hdr_add_ids( T, NxN, R ).

/*
hdr_to_lower( [], [] ).
hdr_to_lower( [H|T], [Hlow|Tlow] ) :-
	( (H =< 0'Z, 0'A =< H) ->
		Hmarg is H - 0'A,
		Hlow  is Hmarg + 0'a
		;
		Hlow = H
	),
	hdr_to_lower( T, Tlow ).
	*/
