:- dynamic( n_table/3 ).

:- use_module(library(terms)).

:- ensure_loaded( '../../src/init_lib' ).
:- pl( sicstus(_S), ensure_loaded( '../../bns/src/llhood/logamma_sicstus' ) ).
:- pl( yap(_Y), reconsult('../../bns/src/llhood/logamma_yap') ).

% Things to check with James.
% Log ? 
% exclusion of constant term ok ?
%
% excluding constant term that is
%
% These assume Alpha_i = 1.
% We follow, Eq. 16, Bayesian Cart Model Search, Chipman et al 
%
cart_llhood( Cart, Llhood ) :-
	cart_llhood_1( Cart, Leaves, Sum ),
	bb_get( knorm, Knorm ),
	Llhood is (Leaves * Knorm) + Sum.

cart_llhood_1( cart(_F,_S,L,R), Lvs, Llhood ) :-
	cart_llhood_1( L, LLvs, LLlhood ),
	cart_llhood_1( R, RLvs, RLlhood ),
	Lvs is LLvs + RLvs,
	Llhood is LLlhood + RLlhood.
cart_llhood_1( leaf(_D,LDIds), 1, Mterm ) :-
	term_hash( LDIds, Hash ),
	( n_table( Hash, LDIds, Mterm ) ->
		true
		; 
		bb_get( all_categories, Kats ),
		bb_get( num_of_categories, K ),
		sum_of_lgammas_n_ik( Kats, LDIds, 0, LgGofNikS ),
		length( LDIds, Ni ),
		NiAddK is Ni + K,
		logamma( NiAddK, LgGofNiAddK ),
		Mterm is LgGofNikS - LgGofNiAddK,
		assert( n_table(LDIds,Mterm) )
	).

sum_of_lgammas_n_ik( [], _LDIds, LgGofNikS, LgGofNikS ).
sum_of_lgammas_n_ik( [Kat|Kats], LDIds, Acc, LgGofNikS ) :-
	findall( 1, (member(Id,LDIds),has_category(Id,Kat)), Ones ),
	length( Ones, N_ik ),
	N_ik_Inc is N_ik + 1,
	logamma( N_ik_Inc, LgGofN_ik_Inc ),
	NxAcc is Acc + LgGofN_ik_Inc,
	sum_of_lgammas_n_ik( Kats, LDIds, NxAcc, LgGofNikS ).

/*
prod_of_gammas_n_ik( [], _, PofGNik, PofGNik ).
prod_of_gammas_n_ik( [Kat|TKats], LDIds, Acc, PofGNik ) :-
	findall( 1, has_category(_Id,Kat), Ones ),
	length( Ones, Nik ),
	gamma( Nik, GNik ),
	NxAcc is Acc * GNik ,
	prod_of_gammas_n_ik( TKats, LDIds, NxAcc, PofGNik ).
	*/
