:- ensure_loaded( '../../src/init_lib' ).
:- requires( sys_info/3 ).
:- elm( '/bns/src/llhood/lhood_canonical_bdeu_multi_spl' ).
:- ensure_loaded( true_abbrv_codes_to_true_name ).

chain_to_bst_mst( [ChainFile] ) :-
	atom_concat( 'gunzip ', ChainFile, Gunzip ),
	sys_info( Gunzip ),
	atom_concat( Stem, '.gz', ChainFile ),
	stem_loads_data( Stem ),
	open( Stem, read, In ),
	read( In, InTerm ),
	chain_to_s_tuples( InTerm, In, [], Stuples ),
	write( done_stuples ), nl,
	close( In ),
	garbage_collect,
	garbage_collect_atoms,
	scan_s_tuples_for_bst_mst( Stuples, Bst, Mst ),
	write( done_bst_mst ), nl,
	% write( done_bst_mst(Bst,Mst) ), nl,
	atom_concat( Stem, '_bst_mst.pl', SFile ),
	open( SFile, write, Out ),
	writeq( Out, Bst ), write( Out, '.' ), nl( Out ),
	writeq( Out, Mst ), write( Out, '.' ), nl( Out ),
	close( Out ),
	atom_concat( 'gzip ', Stem, Gzip ),
	sys_info( Gzip ).
	
scan_s_tuples_for_bst_mst( [s(Mod,LgL,Tms)|T], Bst, Mst ) :-
	scan_s_tuples_for_bst_mst_1( T, LgL, Mod/Tms, Tms, Mod/LgL, Bst, Mst ).

scan_s_tuples_for_bst_mst_1( [], BLgL, BMod/BTms, MTms, MMod/MLgL, Bst, Mst ) :-
	Bst = bst_lgllhood(BMod,BLgL,BTms),
	Mst = mst_visited(MMod,MLgL,MTms).
scan_s_tuples_for_bst_mst_1( [s(HMod,HLgL,HTms)|T], LgL, LMod, Tms, TMod, Bst, Mst ) :-
	( HLgL > LgL ->
		NxLgL = HLgL, NxLMod = HMod/HTms
		;
		NxLgL = LgL, NxLMod = LMod
	),
	( HTms > Tms -> 
		NxTms = HTms, NxTMod = HMod/HLgL
		;
		NxTms = Tms, NxTMod = TMod
	),
	scan_s_tuples_for_bst_mst_1( T, NxLgL, NxLMod, NxTms, NxTMod, Bst, Mst ).

chain_to_s_tuples( end_of_file, _In, Stuples, Stuples ) :-
	!.
chain_to_s_tuples( c(Mod), In, AccTuples, Stuples ) :-
	!,
	( seen_in_acc_tuples(AccTuples,Mod,NxtTuples) ->
		true
		;
		model_llhood( Mod, ModContrs ),
		model_contributions_summation( ModContrs, 0, ModLgLlh ),
		NxtTuples = [s(Mod,ModLgLlh,1)|AccTuples]
	),
	read( In, NxtTerm ),
	chain_to_s_tuples( NxtTerm, In, NxtTuples, Stuples ).
chain_to_s_tuples( _, In, AccTuples, Stuples ) :-
	read( In, NxtTerm ),
	chain_to_s_tuples( NxtTerm, In, AccTuples, Stuples ).
	
model_contributions_summation( [], ModLgLlh, ModLgLlh ).
model_contributions_summation( [_F+FContr|T], Acc, ModLgLlh ) :-
	NxtAcc is Acc + FContr,
	model_contributions_summation( T, NxtAcc, ModLgLlh ).

seen_in_acc_tuples( [s(Mod,LgL,Tms)|T], Mod, NxtTuples ) :-
	!,
	IncTms is Tms + 1,
	NxtTuples = [s(Mod,LgL,IncTms)|T].
seen_in_acc_tuples( [H|T], Mod, [H|TNxtTuples] ) :-
	seen_in_acc_tuples( T, Mod, TNxtTuples ).

stem_loads_data( Stem ) :-
	atom_codes( Stem, StemCs ),
	append( "cm_t", [C1,C2,0'_,0'd,D1,D2,D3|_], StemCs ),
	!,
	once( true_abbrv_codes_to_true_name([C1,C2],True) ),
	( ([D1,D2,D3]=="500") -> 
		DSz = '_500'
		;
		( [D1,D2,D3]=="1K_" ->
			DSz = '_1K'
			;
			atom_codes( RogueDSz, [D1,D2,D3] ),
			throw( unknow_data_size(RogueDSz) )
		)
	),
	atom_concat( True, DSz, DataBase ),
	atom_concat( '../data/', DataBase, DataFile ),
	ensure_loaded( DataFile ),
	atom_concat( True, '_vrs', TrueVrs ),
	atom_concat( '../true/', TrueVrs, TrueVrsFile ),
	ensure_loaded( TrueVrsFile ).
