:- ensure_loaded( library(lists) ). 	% append/3, reverse/2.
:- ensure_loaded( library(system) ). 	% datime/1
%
% expected_loss( [LF|LFs], File ) :-
% compute the expected loss  ( = (1/SampleSize) * Sum (LF(ModI))
% where LF is the loss function of a model. LF/2 with 
% LF( ModI, LossOfModI ) should be defined.
% The ModI are taken in the order they appear in file File.
% Each c/1 term, in this file is assumed to hold a model.
% Other terms in the file are ignored.
%
expected_loss( [], _Skp, _Dir, _File, [] ).
expected_loss( [L|Ls], Skp, Dir, File, [LFile|TLFiles] ) :-
	el_basename( File, Base ),
	atom_concat( Dir, '/', DirSl ),
	atom_concat( DirSl, Base, LFileStem ),
	loss_id( L, Lid ),
	atom_concat( '.l', Lid, LExt ),
	atom_concat( '.m', Lid, MExt ),
	atom_concat( LFileStem, LExt, LFile ),
	atom_concat( LFileStem, MExt, MFile ),
	datime( Start ),
	writeq( starting_loss(L,Start) ), write( '.' ), nl,
	open( File, read, In ),
	open( LFile, write, Out ),
	read( In, Term ),
	loss_sum( Term, none__/_, In, Out, L, 0, 0, Skp, null, null, Sum, Items, N, X ),
	close( In ),
	close( Out ),
	ExpL is Sum / (Items-Skp),
	atom_concat( '.e', Lid, EExt ),
	atom_concat( LFileStem, EExt, ExpLF ),
	open( ExpLF, write, W ),
	writeq( W, expected(ExpL,Sum/(Items-Skp)) ), write( W, '.' ), nl( W ),
	close( W ),
	open( MFile, write, MOut ),
	writeq( MOut, min_max(N,X) ), write( MOut, '.' ), nl( MOut ),
	close( MOut ),
	datime( End ),
	writeq( end_of_loss_at(L,End) ), write( '.' ), nl,
	expected_loss( Ls, Skp, Dir, File, TLFiles ).

expected_loss( Ls, Skp, Dir, File ) :-
     expected_loss( Ls, Skp, Dir, File, _Files ).

loss_sum( end_of_file, _Last, _In, _Out, _L,  Sum, Items, _Skp, N, X, Sum, Items, N, X ) :- !.
loss_sum( Term, Last/LL, In, Out, L, AccS, AccI, Skp, An, Ax, Sum, Items, N, X ) :-
	% 20050711: Untested changes, which are expected to work fine.
	( Term = c(C) -> 
		NxI is AccI + 1,
		( C == Last ->
			ThisLoss = LL
			;
			model_to_loss_structure( C, [], CompC ),
			Call =.. [L,CompC,ThisLoss],
			call( Call )
		),
		writeq( Out, l(ThisLoss) ), write( Out, '.' ), nl( Out ),
		( Skp < NxI -> 
			NxS is AccS + ThisLoss
			;
			NxS is AccS
		),
		update_min_max( An, Ax, ThisLoss, Nn, Nx ),
		NxLast = C, 
		NxLL   = ThisLoss
		;
		NxI is AccI,
		NxS is AccS,
		Nn = An, 
		Nx = Ax,
		NxLast = Last, 
		NxLL   = LL
	),
	read( In, NxTerm ),
	!,
	loss_sum( NxTerm, NxLast/NxLL, In, Out, L, NxS, NxI, Skp, Nn, Nx, Sum, Items, N, X ).

update_min_max( null, null, This, This, This ) :-
	!.
update_min_max( An, Ax, This, Nn, Nx ) :-
	( This < An -> Nn is This, Nx is Ax ; 
		Nn is An,
		( This > Ax -> Nx is This ; Nx is Ax ) ).

el_basename( File, Base ) :-
	atom_codes( File, FileCs ),
	reverse( FileCs, ElifCs ),
	( append( EsabCs, [0'/|Tail], ElifCs ) ->
		( EsabCs==[] ->
			ToRev = Tail
			;
			ToRev = EsabCs
		),
		reverse( ToRev, BaseCs ),
		atom_codes( Base, BaseCs )
		;
		Base = File
	).

model_to_loss_structure( [], UnsLStr, LStr ) :-
	sort( UnsLStr, LStr ).
model_to_loss_structure( [Ch-Pa|T], Acc, LStr ) :-
	parents_child_add_loss_edges( Pa, Ch, Acc, NxAcc ),
	model_to_loss_structure( T, NxAcc, LStr ).

parents_child_add_loss_edges( [], _Ch, Out, Out ).
parents_child_add_loss_edges( [H|T], Ch, In, Out ) :-
	% ord_add_element( In, H-Ch, Mid ),
	parents_child_add_loss_edges( T, Ch, [H-Ch|In], Out ).
