
:- ensure_loaded( library(system) ).    % system/1, environ/2.
:- ensure_loaded( library(swi) ).       % flatten/2.

/* seeds in our publications : 101, 104, 108, 103, 107.

     run_mix( Set, Seed, FromFold, ToFold ).
     Set in {'A','B','C'} or {a,b,c}
     1 =< Seed =< 109  
     1 =< FromFold =< 10
     1 =<   ToFold =< 10

     test: 

     run_mix( b, 101, 1, 2 ).
     
     runs folds 1 and 2 from dataset B with seed 101

     */

run_mix( SetIn, SeedAN, FromAN, ToAN ) :-
     ( (\+var(SetIn),to_set(SetIn,Set)) ->
          true
          ;
          write( user_error, 'Unrecognised dataset ':SetIn ), 
          nl( user_error ),
          write( user_error, 'Please retry.' ),
          nl( user_error ),
          fail
     ),
     to_number( SeedAN, Seed ),
     to_number( FromAN, From ),
     to_number( ToAN, To ),
     write( from(From)-to(To) ), nl,
     run_mix_loop( From, To, Set, Seed ).

run_mix_loop( I, Lim, _Set, _Seed ) :- I > Lim, !.
run_mix_loop( I, Lim, Set,  Seed ) :-
     write( doing(I) ), nl,
     number_codes( I, Ics ), 
     min_two_digit_number_codes( Ics, TwoCs ),
     number_codes( Seed, SeedCs ),
     flatten( ["pyk_dset_",Set,"_g",TwoCs,"_mix_s",SeedCs], PredCs ),
     atom_codes( Pred, PredCs ),
     create_mcmcms_file( Pred ),
     create_runs_file( Pred, Set, Seed, TwoCs ),
     % atom_concat( 'upsh f ', Pred, Upsh ), 
     yap_exec( YapExec ),
     atom_concat( YapExec, Pred, SysCall ), 
     % atom_concat( 'yap -f -l ', Pred, SysCall ), 
     write( sys_call(SysCall) ), nl,
     system( SysCall ),
     NxI is I + 1,
     run_mix_loop( NxI, Lim, Set, Seed ).

create_mcmcms_file( Pred ) :-
     atom_concat( Pred, '.pl', PredPl ),
     open( PredPl, write, Out ),
     nl( Out ),
     portray_clause( Out, (:- ensure_loaded( '../../../run' )) ),
     portray_clause( Out, (:- elm( 'auxil/create_mvisits' )) ),
     nl( Out ),
     ( environ(mcr,McrIn) -> atom_concat(McrIn,'/',Mcr) ; Mcr = 'res/' ),
     % atom_concat( '/usr/people/nicos/res/carts/dsets/', Pred, Dir ),
     % atom_concat( '/usr/people/nicos/res/carts/dsets/', Pred, Dir ),
     atom_concat( Mcr, Pred, Dir ),
     atom_concat( 'runs/runs_', Pred, RunsF ),
     Body = ( Runs=[model_type(cart_red(20)),runs_file(RunsF),results_dir(Dir),print([28])],run(Runs) ),
     portray_clause( Out, (Pred :- Body) ),
     nl( Out ),
     portray_clause( Out, (:- Pred, halt) ), nl( Out ),
     close( Out ).

create_runs_file( Pred, Set, Seed, TwoCs ) :-
     atom_concat( 'runs/runs_', Pred, RunsF ),
     atom_concat( RunsF, '.pl', RunsFPl ),
     open( RunsFPl, write, Out ),
     nl( Out ),
     portray_clause( Out, (:- use_module(library(swi))) ),
     nl( Out ),
     Head = run_data(I,uc,ids_dyn,rm,Data,SGl,OutF,100000,[],SdIdx,all),
     flatten( ["pyk_dset_",Set,0'_, TwoCs], PykCs ),
     % atom_codes( Pyk, [0'p,0'y,0'k,0'_,0's,0'|TwoCs] ),
     atom_codes( Pyk, PykCs ),
     Body = 
       (  Data = idsd_fprs( Pyk, Ids, Lim ),
          SGl = cart/[Ids,Lim,0.95/1],
          OutF = a0_95b1,
          between( 1, 100, I ),
          SdIdx is Seed + (I-1) * 9
       ),
     portray_clause( Out, (Head :- Body) ),
     close( Out ).

yap_exec( YapExec ) :-
     % atom_concat( '/local/usr/nicos/local/bin/yap -f -l ', Pred, SysCall ), 
     library_directory( Lib ),
     atom_concat( Stem, '/share/Yap', Lib ),
     !,
     atom_concat( Stem, '/bin/yap -f -l ', YapExec ).

to_set( a, "A" ).
to_set( 'A', "A" ).
to_set( b, "B" ).
to_set( 'B', "B" ).
to_set( c, "C" ).
to_set( 'C', "C" ).

to_number( In, Number ) :-
	( atom(In) -> 
		atom_codes( In, InCs ),
		number_codes( Number, InCs )
		;
		( number(In) ->
			Number = In 
			;
			% In = [_|_],
			number_codes( Number, In )
		)
	).

min_two_digit_number_codes( [One], [0'0,One] ) :- !.
min_two_digit_number_codes( Other, Other ).
